#!/usr/bin/env python

########################################################################

import os
import sys
import random
import time
import pygame
import pygame.mixer

########################################################################

def data_file(f):
    workdir = os.path.split(sys.argv[0])[0]
    result = os.path.join(workdir, "data", f)
    return result

########################################################################

snub_index = 0
snub_speak = [
    ("hey_hello.ogg", "Hey! Hello! Where are you going?"),
    ("come_back.ogg", "Come back!"),
    ("why_avoid.ogg", "Why are you avoiding me?"),
    ("please_dont_run.ogg", "Please don't run away!"),
    ("dont_leave_me.ogg", "Don't leave me here alone!"),
    ("pleeease.ogg", "Please?"),
    ]

def next_snub():
    global snub_index, snub_speak
    say = snub_speak[snub_index]
    snub_index = (snub_index + 1) % len(snub_speak)
    return say

########################################################################

class Screen(object):
    
    def __init__(self):
        self.surf = pygame.display.set_mode((640, 400))
        self.black = pygame.Color("black")
        pygame.display.set_caption("The Snubbing of the Squid")
    
    def clear(self):
        self.surf.fill(self.black)

########################################################################

class Unit(object):

    def __init__(self, map, BrainClass, images, pos=(0.0, 0.0, 0.0)):
        #print "Spawn", images, pos
        self.map = map
        self.brain = BrainClass(self)
        self.images = images
        self.x = float(pos[0])
        self.y = float(pos[1])
        self.z = float(pos[2])
        self.frames = {}
        self.frames_flipped = {}
        self.fr = None
        self.flip = False
        self.climb = False
        self.stomach = []
        self.tick = 0
        self.disgorged = False
        self.ate = False
    
    def clear(self):
        #print "Removing", self
        self.map.units.remove(self)
    
    def get_free_pos(self):
        return (self.x, self.y, self.z)
    
    def get_tile_pos(self):
        return (int(round(self.x)), int(round(self.y)), int(round(self.z)))

    def get_offset_tile_pos(self, offset=(0.0, 0.0, 0.0)):
        (ox, oy, oz) = offset
        return (int(round(self.x + ox)), int(round(self.y + oy)), int(round(self.z + oz)))

    def draw(self, dest_surf, xy):
        if self.fr is None or self.fr not in self.frames:
            self.render()
        (x, y) = xy
        if self.flip:
            framepool = self.frames_flipped
        else:
            framepool = self.frames
        surf = framepool[self.fr]
        cx = x - surf.get_width() / 2 + self.x_offset()
        cy = y - surf.get_height() / 2 + self.y_offset()
        dest_surf.blit(surf, (cx, cy))

    def x_offset(self):
        return 0

    def y_offset(self):
        return 0
        
    def render(self):
        unit_dir = data_file(os.path.join("unit", self.images))
        for filename in os.listdir(unit_dir):
            if filename.endswith(".png"):
                f = os.path.join(unit_dir, filename)
                #print f
                surf = pygame.image.load(f).convert_alpha()
                key = filename[0:-4]
                self.frames[key] = surf
                fsurf = pygame.transform.flip(surf, True, False)
                self.frames_flipped[key] = fsurf

    def handle(self, event):
        self.brain.handle(event)

    def update(self):
        self.tick += 1
        self.brain.update()
        (mx, my, mz) = self.brain.motion()
        mz += self.gravity()
        self.check_collisions((mx, my, mz))
        if mx > 0 or my > 0:
            self.flip = True
        if mx < 0 or my < 0:
            self.flip = False

    def check_collisions(self, try_move):
        "check for collision, and apply motion"
        move = [m for m in try_move]
        self.climb = False
        self.moving = False
        for axis in xrange(3):
            if move[axis] == 0.0:
                continue
            for (cornerx, cornery) in ((-0.25, -0.25), (0.25, -0.25), (0.25, 0.25), (-0.25, 0.25)):
                if move[axis] == 0.0:
                    continue
                offset = [0.0 + cornerx, 0.0 + cornery, 0.0]
                offset[axis] += move[axis]
                pos = self.get_offset_tile_pos(offset)
                block = self.map.get_block(pos)
                if block:
                    move[axis] = 0.0
                    if axis < 2:
                        above = self.map.get_block((pos[0], pos[1], pos[2] + 1))
                        if not above:
                            self.climb = True
                            move[2] += self.climbrate()
        (mx, my, mz) = move
        onblock = self.map.get_block(self.get_tile_pos())
        if onblock:
            mz += self.popout()
        self.x += mx
        self.y += my
        self.z += mz
        self.moving = mx <> 0.0 or my <> 0.0 or mz <> 0.0

    def textbox(self, sfx, words):
        self.map.game.textbox(sfx, words)

    def gravity(self):
        return -0.4

    def speed(self):
        return 0.15

    def popout(self):
        return 0.6

    def climbrate(self):
        return 0.6

    def eat(self):
        "Eats the block the unit is standing on"
        pos = self.get_offset_tile_pos((0.0, 0.0, -1.0))
        block = self.map.pickup_block(pos)
        self.stomach.append(block)
        if block:
            self.ate = self.tick
            self.map.game.sfx("eat.ogg")
    
    def disgorge(self):
        if len(self.stomach) == 0:
            return
        block = self.stomach.pop()
        pos = self.get_tile_pos()
        self.map.place_block(pos, block)
        self.disgorged = self.tick 
        self.map.game.sfx("disgorge.ogg")

#-----------------------------------------------------------------------

class EtherUnit(Unit):
    
    def check_collisions(self, try_move):
        (mx, my, mz) = try_move
        self.x += mx
        self.y += my
        self.z += mz
    
#-----------------------------------------------------------------------

class Enui(Unit):
    
    def __init__(self, map, pos):
        super(Enui, self).__init__(map, PlayerBrain, "enui", pos)
        self.wiggle = 0
    
    def y_offset(self):
        return -10
    
    def update(self):
        super(Enui, self).update()
        frameset = "walk"
        if self.climb:
            frameset = "climb"
        self.wiggle = (self.wiggle + 1) % 6
        self.fr = "%s%s" % (frameset, self.wiggle / 3)
        if self.ate and self.ate > self.tick - 3:
            self.fr = "eat"
        if self.disgorged and self.disgorged > self.tick - 3:
            self.fr = "disgorge"

#-----------------------------------------------------------------------

class Snubthar(EtherUnit):
    
    def __init__(self, map, pos):
        super(Snubthar, self).__init__(map, SnubBrain, "snubthar", pos)
        self.spoketo = False

    def gravity(self):
        if self.spoketo:
            return 0.24
        return 0.08

    def speed(self):
        return 0.02
    
    def update(self):
        super(Snubthar, self).update()
        (tx, ty, tz) = self.get_tile_pos()
        self.fr = "drift"
        p = self.map.player
        if tz > p.z + 1 and tz <= p.z + 2:
            self.fr = "snub"
        if not self.spoketo:
            if abs(tx - p.x) < 2 and abs(ty - p.y) < 2 and abs(tz - p.z) < 4:
                self.textbox(*next_snub())
                self.spoketo = True
        if tz > 35:
            self.clear()

#-----------------------------------------------------------------------

class Totem(Unit):

    def __init__(self, map, pos):
        super(Totem, self).__init__(map, ShyBrain, "totem", pos)
        self.spoketo = False
        self.wiggle = 0

    def speed(self):
        if self.spoketo:
            return 0.25
        return 0.125
    
    def update(self):
        super(Totem, self).update()
        frameset = "walk"
        #if self.climb:
        #    frameset = "climb"
        self.wiggle = (self.wiggle + 1) % 6
        frame = 1
        if self.moving:
            frame = self.wiggle / 3
        self.fr = "%s%s" % (frameset, frame)
        
        (tx, ty, tz) = self.get_tile_pos()
        p = self.map.player
        if not self.spoketo:
            if abs(tx - p.x) < 1 and abs(ty - p.y) < 1 and abs(tz - p.z) < 2:
                self.textbox(*next_snub())
                self.spoketo = True
            

########################################################################

class Brain(object):

    def __init__(self, unit):
        self.unit = unit

    def handle(self, event):
        pass

    def update(self):
        pass

    def motion(self):
        return (0.0, 0.0, 0.0)

    def follow(self, other, move=(0.0, 0.0, 0.0)):
        (x, y, z) = move
        u = self.unit
        sp = u.speed()
        if u.x < other.x:
            x += sp
        if u.x > other.x:
            x -= sp
        if u.y < other.y:
            y += sp
        if u.y > other.y:
            y -= sp
        return (x, y, z)
    
    def avoid(self, other, move=(0.0, 0.0, 0.0)):
        (x, y, z) = move
        u = self.unit
        sp = u.speed()
        if u.x < other.x:
            x -= sp
        if u.x > other.x:
            x += sp
        if u.y < other.y:
            y -= sp
        if u.y > other.y:
            y += sp
        return (x, y, z)

#-----------------------------------------------------------------------

class SnubBrain(Brain):
    
    def __init__(self, unit):
        super(SnubBrain, self).__init__(unit)
        self.first = True
    
    def motion(self):
        u = self.unit
        if self.first:
            self.first = False
            return self.follow(u.map.player)
        (px, py, pz) = u.map.player.get_free_pos()
        if u.z > pz + 2:
            return self.avoid(u.map.player)
        return (0.0, 0.0, 0.0)

#-----------------------------------------------------------------------

class ShyBrain(Brain):
    
    def __init__(self, unit):
        super(ShyBrain, self).__init__(unit)
        self.first = True
        self.spooked = False

    def motion(self):
        u = self.unit
        if self.first:
            self.first = False
            return self.follow(u.map.player)
        p = u.map.player
        if self.spooked or abs(u.x - p.x) < 4 and abs(u.y - p.y) < 4 and abs(u.z - p.z) < 4:
            self.spooked = True
            return self.avoid(p)
        if abs(u.x - p.x) > 8 or abs(u.y - p.y) > 8 or abs(u.z - p.z) > 8:
            self.spooked = False
        if abs(u.x - p.x) > 12 or abs(u.y - p.y) > 12 or abs(u.z - p.z) > 12:
            u.clear()
        return (0.0, 0.0, 0.0)

#-----------------------------------------------------------------------

class PlayerBrain(Brain):

    def __init__(self, unit):
        super(PlayerBrain, self).__init__(unit)
        self.up = False
        self.down = False
        self.left = False
        self.right = False

    def handle(self, event):
        if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_UP:
                self.up = True
            elif event.key == pygame.K_DOWN:
                self.down = True
            elif event.key == pygame.K_LEFT:
                self.left = True
            elif event.key == pygame.K_RIGHT:
                self.right = True
        elif event.type == pygame.KEYUP:
            if event.key == pygame.K_UP:
                self.up = False
            elif event.key == pygame.K_DOWN:
                self.down = False
            elif event.key == pygame.K_LEFT:
                self.left = False
            elif event.key == pygame.K_RIGHT:
                self.right = False
            elif event.key == pygame.K_x:
                self.unit.eat()
            elif event.key == pygame.K_z:
                self.unit.disgorge()

    def update(self):
        self.unit.map.view.follow(self.unit)
    
    def motion(self):
        x = 0.0
        y = 0.0
        sp = self.unit.speed()
        if self.up:
            y -= sp
            x += sp
        if self.down:
            y += sp
            x -= sp
        if self.left:
            x -= sp
            y -= sp
        if self.right:
            x += sp
            y += sp
        return (x, y, 0.0)

#-----------------------------------------------------------------------

########################################################################

class UnitManager(object):
    
    def __init__(self, map):
        self.map = map
        self.data = []
    
    def add(self, u):
        self.data.append(u)
        return u
    
    def remove(self, u):
        self.data.remove(u)

    def tile_pos_cache(self):
        for u in self.data:
            u._tile_pos = u.get_tile_pos()
    
    def should_draw_at(self, pos):
        for u in self.data:
            if u._tile_pos == pos:
                return u
        return None

    def handle(self, event):
        for u in self.data:
            u.handle(event)
    
    def update(self):
        for u in self.data:
            u.update()

########################################################################

class Block(object):
    
    def draw(self, dest_surf, xy):
        if self.surf is None:
            self.render()
        (x, y) = xy
        center_xy = (x - self.surf.get_width() / 2, y - self.surf.get_height() / 2)
        dest_surf.blit(self.surf, center_xy)
        
    def render(self):
        f = data_file(os.path.join("block", self.file()))
        #print f
        self.surf = pygame.image.load(f).convert_alpha()
    
    def opaque(self):
        return True
        
#-----------------------------------------------------------------------

class Sand(Block):

    surf = None

    def file(self):
        return "sand.png"

class PockedSand(Block):

    surf = None

    def file(self):
        return "pockedsand.png"

class X_Sand(Block):

    surf = None

    def file(self):
        return "x_sand.png"

class Glass(Block):
    
    surf = None
    
    def file(self):
        return "glass.png"
    
    def opaque(self):
        return False

########################################################################

class View(object):
    
    def __init__(self):
        self.x = 0.0
        self.y = 0.0
        self.z = 0.0
        self.size = ViewSize()
    
    def pos(self):
        return (self.x, self.y, self.z)

    def handle(self, event):
        self.size.handle(event)
    
    def follow(self, u):
        (self.x, self.y, z) = u.get_free_pos()
        self.z = z *.8

########################################################################

class ViewSize(object):
    
    def __init__(self):
        self.x = 9
        self.y = 9
        self.z = 5

    def z_shrink(self):
        self.z = max(0, self.z - 1)
    
    def z_grow(self):
        self.z = min(10, self.z + 1)

    def handle(self, event):
        if event.type == pygame.KEYUP:
            if event.key == pygame.K_LEFTBRACKET:
                self.z_shrink()
            elif event.key == pygame.K_RIGHTBRACKET:
                self.z_grow()    

########################################################################

class Spawner(object):
    
    def __init__(self, map):
        self.map = map
        self.intervals = {}
        self.set_interval("snubthar", self.spawn_snubthar, 25.0)
        self.set_interval("totem", self.spawn_totem, 123.0)

    def set_interval(self, key, callback, delay):
        self.intervals[key] = (callback, delay, time.time())

    def update(self):
        for key in self.intervals:
            (callback, delay, last) = self.intervals[key]
            if time.time() > last + delay:
                callback()
                self.set_interval(key, callback, delay)

    def beneath(self, pos, dist):
        return (pos[0], pos[1], pos[2] - dist)
    
    def nearby(self, pos, ldist=3, rdist=5):
        result = [i for i in pos]
        for i in xrange(2):
            result[i] += random.randint(ldist, rdist) * random.choice((-1, 1))
        return result
    
    def h_nearby(self, pos):
        result = [i for i in pos]
        result[0] += random.randint(4,6) * random.choice((-1, 1))
        return result
    
    def spawn_snubthar(self):
        m = self.map
        where = self.beneath(self.h_nearby(m.view.pos()), 2)
        m.units.add(Snubthar(m, where))

    def spawn_totem(self):
        m = self.map
        where = self.nearby(m.view.pos(), 8, 10)
        m.units.add(Totem(m, where))

########################################################################

class Map(object):
    
    def __init__(self, game, seed=None):
        self.game = game
        self.terrain = Terrain(seed)
        self.data = {}
        self.data[(0,0,-1)] = X_Sand()
        self.view = View()
        self.units = UnitManager(self)
        self.player = self.units.add(Enui(self, (0,0,0)))
        self.spawner = Spawner(self)
        self.depther = pygame.Surface(self.game.screen.surf.get_size()).convert_alpha()
        self.depther.fill(pygame.Color("#00000010"))
        
    def update(self):
        self.spawner.update()
        self.units.update()
    
    def draw(self, surf):
        covered = []
        queue = []
        count = 0
        self.units.tile_pos_cache()
        layer_used = False
        for pos in self.front_to_back(self.view.pos()):
            if pos == "layer":
                if layer_used:
                    queue.append("layer")
                    layer_used = False
            else:
                u = self.units.should_draw_at(pos)
                xy = self.screen_xy(surf, pos)
                b = None
                if xy not in covered:
                    b = self.get_block(pos)
                if u or b:
                    queue.append((u, b, xy))
                    layer_used = True
                    if b and b.opaque():
                        covered.append(xy)
        for element in reversed(queue):
            if element == "layer":
                surf.blit(self.depther, (0,0))
            else:
                (u, b, xy) = element
                if b:
                    b.draw(surf, xy)
                    count += 1
                if u:
                    u.draw(surf, self.screen_xy(surf, u.get_free_pos()))
    
    def place_block(self, pos, block):
        old = self.get_block(pos)
        if old:
            pass
            #print "block was in the way", old
        else:
            self.data[pos] = block
    
    def pickup_block(self, pos):
        block = self.get_block(pos)
        self.data[pos] = None
        return block
    
    def get_block(self, pos):
        if not pos in self.data:
            self.default_block(pos)
        return self.data[pos]
    
    def default_block(self, pos):
        self.data[pos] = self.terrain.gen(pos)
    
    def screen_xy(self, screen_surf, pos):
        centerx = int(screen_surf.get_width() / 2)
        centery = int(screen_surf.get_height() / 2)
        vp = self.view.pos()
        (ax, ay, az) = [pos[i] - vp[i] for i in xrange(3)]
        # calc screen positions
        x = centerx + ax * 32 + ay * 32
        y = centery - ax * 18 + ay * 18 - az * 36
        return (x, y)
    
    def back_to_front(self, origin):
        size = self.view.size
        (ox, oy, oz) = origin
        for z in xrange(oz - size.z, oz + size.z + 1):
            for y in xrange(oy - size.y, oy + size.y + 1):
                for x in xrange(ox + size.x, ox - size.x - 1, -1):
                    yield (x, y, z)

    def front_to_back(self, origin):
        size = self.view.size
        (ox, oy, oz) = [int(round(o)) for o in origin]
        for z in xrange(oz + size.z, oz - size.z - 1, -1):
            for y in xrange(oy + size.y, oy - size.y - 1, -1):
                for x in xrange(ox - size.x, ox + size.x + 1, 1):
                    yield (x, y, z)
            yield "layer"

    def handle(self, event):
        if event.type == pygame.KEYUP:
            if event.key == pygame.K_s:
                self.spawner.spawn_snubthar()
            if event.key == pygame.K_t:
                self.spawner.spawn_totem()
        self.view.handle(event)
        self.units.handle(event)
    
########################################################################

class Terrain(object):
    
    def __init__(self, seed=None):
        if seed is None:
            self.seed = random.randint(-2000000000, 2000000000)
        else:
            self.seed = seed
        self.sand = Sand()
        self.pockedsand = PockedSand()
        self.glass = Glass()
    
    def randint(self, lbound, rbound, seedadd=0):
        #state = random.getstate()
        random.seed(self.seed + seedadd)
        result = random.randint(lbound, rbound)
        #random.setstate(state)
        return result
    
    def gen(self, pos):
        (x, y, z) = pos
        seedadd = x + y * 10000000
        fl = self.randint(-1, 0, seedadd)
        if self.randint(0, 99, seedadd) < 2:
            gl = self.randint(2, 4, seedadd)
            if z >= fl and z < gl:
                return self.glass
        if z < fl:
            r = self.randint(0, 99, seedadd)
            if r <= 10:
                return self.pockedsand
            return self.sand

########################################################################

class TextBox(object):
    
    def __init__(self, game, sfx, words):
        game.sfx(sfx)
        self.game = game
        self.words = words
        self.surf = None
        font_file = data_file(os.path.join("font", "TimesGothic.ttf"))
        self.font = pygame.font.Font(font_file, 64)
        self.color = pygame.Color("white")
        self.age = time.time()
    
    def clear(self):
        self.surf = None
        self.game.text = None
        self.game = None
    
    def draw(self, dest_surf):
        if not self.surf:
            self.render(dest_surf)
        dest_surf.blit(self.surf, (0,0))

    def render(self, dest_surf):
        line_padding = 0
        word_padding = 20
        surf = pygame.surface.Surface(dest_surf.get_size()).convert_alpha()
        surf.fill(pygame.Color("#00000000"))
        where = [0,0]
        line_height = 0
        for line in self.words.split("\n"):
            for word in line.split(" "):
                tsurf = self.font.render(word, True, self.color).convert_alpha()
                if where[0] + tsurf.get_width() > surf.get_width():
                    where[0] = 0
                    where[1] += line_height + line_padding
                    line_height = 0
                surf.blit(tsurf, where)
                where[0] += tsurf.get_width() + word_padding
                line_height = max(line_height, tsurf.get_height())
            where[0] = 0
            where[1] += line_height + line_padding
        self.surf = surf

    def handle(self, event):
        pass
    
    def update(self):
        if time.time() > self.age + len(self.words) / 6:
            self.clear()

########################################################################

class Hud(object):
    
    def __init__(self, game):
        self.game = game
        backdrop_filename = data_file(os.path.join("hud", "hud.png"))
        self.backdrop = pygame.image.load(backdrop_filename).convert_alpha()
        font_file = data_file(os.path.join("font", "FlagGothic.ttf"))
        self.font = pygame.font.Font(font_file, 56)
        self.color = pygame.Color("white")
        self._stomach = None
        self.surfaces = {}
    
    def update(self):
        size = len(self.game.map.player.stomach)
        if self._stomach <> size:
            self.surfaces["stomach"] = str(size)
        
    def draw(self, dest_surf):
        dest_surf.blit(self.backdrop, (0,0))
        if self.game.text is None:
            for key, val in self.surfaces.iteritems():
                if isinstance(val, basestring):
                    self.render(key, val, dest_surf)
            for key in self.surfaces:
                (surf, xy) = self.surfaces[key]
                dest_surf.blit(surf, xy)

    def render(self, key, txt, dest_surf):
        surf = self.font.render(txt, True, self.color)
        xy = (0, 0)
        self.surfaces["stomach"] = (surf, xy)
        

########################################################################

class Victory(object):
    
    def __init__(self, game):
        self.game = game
        self.frames = []
        for i in xrange(4):
            filename = data_file(os.path.join("story", "end%d.png" % (i)))
            surf = pygame.image.load(filename)
            self.frames.append(surf)
        self.index = -1
        self.last = time.time()
        self.wait = 1.0
        self.done = False
    
    def update(self):
        t = time.time()
        if t > self.last + self.wait:
            if self.index < 3:
                self.set_index(self.index + 1)
            else:
                self.done = True

    def set_index(self, n):
        self.index = n
        self.last = time.time()
        self.wait = self.game.sfx("end%d.ogg" % (n))
    
    def draw(self, dest_surf):
        if self.index >= 0:
            surf = self.frames[self.index]
            dest_surf.blit(surf, (0,0))
    
    def handle(self, event):
        if self.done:
            if event.type == pygame.KEYDOWN:
                self.game.keep_playing = False

########################################################################

class Game(object):
    
    def __init__(self):
        pygame.init()
        pygame.mixer.init()
        self._sfx = {}
        self.screen = Screen()
        self.clock = pygame.time.Clock()
        self.fps = 32
        self.keep_playing = True
        self.map = Map(self)
        self.text = None
        self.hud = Hud(self)
        self.textbox("i_am_alone.ogg", "Where is everybody?\nI am alone!")
        self.victory = None
    
    def sfx(self, name):
        if not name in self._sfx:
            sfx_file = data_file(os.path.join("sfx", name))
            self._sfx[name] = pygame.mixer.Sound(sfx_file)
        sound = self._sfx[name]
        sound.play()
        return sound.get_length()
    
    def cleanup(self):
        pygame.quit()
    
    def play(self):
        while self.keep_playing:
            for event in pygame.event.get():
                self.handle(event)
            self.update()
            self.draw()
            pygame.display.update()
            self.clock.tick(self.fps)

    def update(self):
        if self.victory:
            self.victory.update()
            return
        self.map.update()
        if self.text:
            self.text.update()
        self.hud.update()
        if self.map.player.z >= 25:
            self.victory = Victory(self)
    
    def draw(self):
        self.screen.clear()
        dest_surf = self.screen.surf
        if self.victory:
            self.victory.draw(dest_surf)
            return
        self.map.draw(dest_surf)
        self.hud.draw(dest_surf)
        if self.text:
            self.text.draw(dest_surf)

    def handle(self, event):
        if event.type == pygame.QUIT:
            self.keep_playing = False
        if self.victory:
            self.victory.handle(event)
            return
        if event.type == pygame.KEYUP:
            if event.key == pygame.K_F9:
                self.victory = Victory(self)
        self.map.handle(event)
        if self.text:
            self.text.handle(event)

    def textbox(self, sfx, words):
        if self.text:
            self.text.clear()
        self.text = TextBox(self, sfx, words)

########################################################################

if __name__ == "__main__":
    game = Game()
    if False:
        import cProfile as profile
        profile.run('game.play()', "profile")
        import pstats
        p = pstats.Stats('profile')
        p.sort_stats('cumulative').print_stats(10)
        p.sort_stats('time').print_stats(10)
    else:
        game.play()
    game.cleanup()
